//---------------------------------------------------------------------------

#include <vcl.h>
#include <Registry.hpp>
#pragma hdrstop

#include "Unit1.h"
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
#define BURST_SIZE 4096
#define STX 2
#define ETX 3

const int baudrates[12] = {
	9600, 19200, 38400, 57600, 115200, 230400, 460800, 921600, 1843200,
	312500, 625000, 1250000 };

const int distance_type[] = { 25, 30, 85, 350, 500, 2000};
const int range_type[] =    {  1,  5, 20, 100, 200,  500};

__fastcall TForm1::TForm1(TComponent* Owner)
	: TForm(Owner)
{
	Application->OnException = AppException;
	burst_buffer = new BYTE[BURST_SIZE];
	handle = INVALID_HANDLE_VALUE;
	timeout = 150;
	base_dist = 0.0, range = 0.0;
}

__fastcall TForm1::~TForm1()
{
	delete [] burst_buffer;
}

void __fastcall TForm1::AppException(TObject *Sender, Exception *E)
{
	Form1->ErrLabel->Font->Color = clFuchsia;
	Form1->ErrLabel->Caption = E->Message;
}

//---------------------------------------------------------------------------
void __fastcall TForm1::PortComboDropDown(TObject *Sender)
{
	AnsiString saved_text = PortCombo->Text;
	PortCombo->Clear();

	TStringList *ss = new TStringList;
	TRegistry *reg = new TRegistry;
	try {
		reg->RootKey = HKEY_LOCAL_MACHINE;
		if (reg->OpenKey("HARDWARE\\DEVICEMAP\\SERIALCOMM", false)) {
			reg->GetValueNames(ss);
			for (int i = 0; i < ss->Count; i++)
				if (reg->GetDataType(ss->Strings[i]) == rdString)
					PortCombo->Items->Add(reg->ReadString(ss->Strings[i]));
		}
		PortCombo->Text = saved_text;
	} __finally {
		delete reg;
		delete ss;
	}
}

void __fastcall TForm1::OpenPortExecute(TObject *Sender)
{
	if (OpenPort->Checked)
		try {
			handle = CreateFile(PortCombo->Text.c_str(), GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
			if (handle == INVALID_HANDLE_VALUE)
				throw Exception("serial port cannot open");

			char *ep;
			double baud = strtod(BaudCombo->Text.c_str(), &ep);
			if (*ep == 'k')
				baud *= 1000;

			DCB dcb;
			GetCommState(handle, &dcb);
			dcb.BaudRate = baud;
			dcb.ByteSize = 8;
			dcb.Parity = NOPARITY;
			dcb.StopBits = ONESTOPBIT;
			SetCommState(handle, &dcb);
		} catch (...) {
			OpenPort->Checked = false;
			throw;
		}
	else {
		CloseHandle(handle);
		handle = INVALID_HANDLE_VALUE;
	}
}

void __fastcall TForm1::BaudComboClick(TObject *Sender)
{
	AnsiString s = BaudCombo->Text.c_str();
	char *ep, *p = s.c_str();
	double baud = strtod(p, &ep);
	if (*ep == 'k' || *ep == 'K')
		baud *= 1000;

	DCB dcb;
	GetCommState(handle, &dcb);
	dcb.DCBlength = sizeof(dcb);
	dcb.BaudRate = (int)baud;
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	SetCommState(handle, &dcb);

	PurgeComm(handle, PURGE_RXCLEAR|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_TXABORT);
}

void __fastcall TForm1::PurgeButtonClick(TObject *Sender)
{
	PurgeComm(handle, PURGE_RXABORT+PURGE_RXCLEAR);
}

//---------------------------------------------------------------------------

void __fastcall TForm1::PortOpenUpdate(TObject *Sender)
{
	TAction *action = dynamic_cast<TAction *>(Sender);
	action->Enabled = (handle != INVALID_HANDLE_VALUE);
}

AnsiString __fastcall stat_str(DWORD port_errors)
{
	AnsiString e = "-----";
	if (port_errors&CE_BREAK)
		e[1] = 'b';
	if (port_errors&CE_FRAME)
		e[2] = 'f';
	if (port_errors&CE_OVERRUN)
		e[3] = 'o';
	if (port_errors&CE_RXOVER)
		e[4] = 'O';
	return e;
}

void __fastcall TForm1::GetTypeExecute(TObject *Sender)
{
	BaudComboClick(NULL);

	const BYTE cmd[5] = "\x02p?\x03\x4C";	// GET_DIST_TYPE
	DWORD written;
	if (!WriteFile(handle, cmd, 5, &written, NULL))
		throw Exception("WriteFile() error");

	DWORD tc0 = GetTickCount();
	for (;;) {
		if (!ClearCommError(handle, &port_errors, &comstat))
			throw Exception("ClearCommError()");
		CommStatLabel->Caption = stat_str(port_errors);

		if (comstat.cbInQue >= 6)
			break;
		if (GetTickCount()-tc0 > timeout)
			throw ECD5Comm("read timeout");
	}

	DWORD read;
	BYTE reply[6];
	if (!ReadFile(handle, reply, 6, &read, NULL))
		throw Exception("read error");

	if (reply[0] != 0x02)
		throw Exception("no STX");
	if (reply[4] != 0x03)
		throw Exception("no ETX");
	if ((reply[1] ^ reply[2] ^ reply[3] ^ 0x03) != reply[5])
		throw Exception("check sum missmatched");

	int t = reply[1]-'0';
	if (t < 0 || t >= 5)
		throw Exception("unknown type");
	//AnsiString s = AnsiString().sprintf("%dmm+-%d", distance_type[t], range_type[t]);
	//AnsiString s = AnsiString().sprintf("%d-%dmm", distance_type[t]-range_type[t], distance_type[t]+range_type[t]);
	TypeCombo->ItemIndex = t;
	base_dist = distance_type[t];
	range = range_type[t];
}

void __fastcall TForm1::Measure1Execute(TObject *Sender)
{
	BaudComboClick(NULL);

	const BYTE cmd[5] = "\x02M?\x03\x71";	// MEASURE1
	DWORD written;
	if (!WriteFile(handle, cmd, 5, &written, NULL))
		throw Exception("WriteFile() error");

	DWORD tc0 = GetTickCount();
	for (;;) {
		if (GetTickCount()-tc0 > tc0+timeout)
			throw ECD5Comm("read timeout");

		if (!ClearCommError(handle, &port_errors, &comstat))
			throw Exception("ClearCommError()");
		CommStatLabel->Caption = stat_str(port_errors);

		if (comstat.cbInQue >= 6)
			break;
	}

	DWORD read;
	BYTE reply[6];
	if (!ReadFile(handle, reply, 6, &read, NULL))
		throw Exception("read error");

	if (reply[0] != 0x02)
		throw Exception("no STX");
	if (reply[4] != 0x03)
		throw Exception("no ETX");
	if ((reply[1] ^ reply[2] ^ reply[3] ^ 0x03) != reply[5])
		throw Exception("check sum missmatched");

	int d = (reply[1]<<16) + (reply[2]<<8) + reply[3];
	DisplEdit->Text = IntToHex(d, 6);
	if (base_dist > 0.0)
		DisplEdit->Text = DisplEdit->Text +"," + AnsiString().sprintf("%8.3fmm", base_dist + range*(d - 0x100000)/0xAAAAA);
}

void __fastcall TForm1::StartContinuousExecute(TObject *Sender)
{
	BaudComboClick(NULL);

	const BYTE cmd[5] = "\x02M1\x03\x7F";	// CONTINUOUS
	DWORD written;
	if (!WriteFile(handle, cmd, 5, &written, NULL))
		throw Exception("WriteFile() error");
	// THIS COMMAND HAS NO REPLY
}

void __fastcall TForm1::StopContinuousExecute(TObject *Sender)
{
	BaudComboClick(NULL);

	const BYTE cmd[5] = "\x02M0\x03\x7E";	// STOP
	DWORD written;
	if (!WriteFile(handle, cmd, 5, &written, NULL))
		throw Exception("WriteFile() error");
	// THIS COMMAND HAS NO REPLY
}


void __fastcall TForm1::SetBaudExecute(TObject *Sender)
{
	BaudComboClick(NULL);

	int b = CD5BaudCombo->ItemIndex;
	BYTE cmd[5];
	cmd[0] = STX;
	cmd[1] = 'B';	// SET_BAUD
	cmd[2] = (b < 10) ? '0'+b : 'A'+b-10;
	cmd[3] = ETX;
	cmd[4] = cmd[1]^cmd[2]^cmd[3];

	DWORD written;
	if (!WriteFile(handle, cmd, 5, &written, NULL))
		throw Exception("WriteFile() error");
	DWORD tc0 = GetTickCount();
	for (;;) {
		if (GetTickCount() > tc0+timeout)
			throw Exception("reply timeout");
		if (!ClearCommError(handle, &port_errors, &comstat))
			throw Exception("ClearCommError()");
		if (comstat.cbInQue >= 6)
			break;
	}
	DWORD read;
	BYTE reply[6];
	if (!ReadFile(handle, reply, 6, &read, NULL))
		throw Exception("read error");

	if (reply[0] != 0x02)
		throw Exception("no STX");
	if (reply[4] != 0x03)
		throw Exception("no ETX");
	if ((reply[1] ^ reply[2] ^ reply[3] ^ 0x03) != reply[5])
		throw Exception("check sum missmatched");
	if (reply[1] == '>') {
		ErrLabel->Font->Color = clNavy;
		ErrLabel->Caption = "ACK";
	} else if (reply[1] == '?') {
		ErrLabel->Font->Color = clFuchsia;
		ErrLabel->Caption = "NAK";
	} else {
		ErrLabel->Font->Color = clFuchsia;
		ErrLabel->Caption = "unknown reply";
	}
}

void __fastcall TForm1::ActiveSweepExecute(TObject *Sender)
{
	Screen->Cursor = crHourGlass;

	DCB dcb;
	int ret = GetCommState(handle, &dcb);
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	bool found = false;
	for (int b = 11; b  >= 0 && !found; b--)
		try {
			dcb.BaudRate = baudrates[b];
			SetCommState(handle, &dcb);

			const BYTE cmd[5] = "\x02p?\x03\x4C";	// GET_DIST_TYPE
			DWORD written;
			if (!WriteFile(handle, cmd, 5, &written, NULL))
				throw Exception("WriteFile() error");

			DWORD tc0 = GetTickCount();
			for (;;) {
				if (!ClearCommError(handle, &port_errors, &comstat))
					throw Exception("ClearCommError()");
				CommStatLabel->Caption = stat_str(port_errors);

				if (comstat.cbInQue >= 6)
					break;
				DWORD tc = GetTickCount();
				if (tc > tc0+timeout)
					throw ECD5Comm("read timeout");
			}

			DWORD read;
			BYTE reply[6];
			if (!ReadFile(handle, reply, 6, &read, NULL))
				throw Exception("read error");

			if (reply[0] != STX)
				throw Exception("no STX");
			if (reply[4] != ETX)
				throw Exception("no ETX");
			if ((reply[1] ^ reply[2] ^ reply[3] ^ 0x03) != reply[5])
				throw Exception("check sum missmatched");

			found = true;
			//BaudEdit->Text = IntToStr((int)dcb.BaudRate);
			BaudCombo->ItemIndex = b;
			ErrLabel->Font->Color = clNavy;
			ErrLabel->Caption = "SUCCEED; baudrate"+IntToStr(dcb.BaudRate);
		} catch (Exception &e) {
			ErrLabel->Font->Color = clFuchsia;
			ErrLabel->Caption = e.Message;
		}
	Screen->Cursor = crDefault;
}
//---------------------------------------------------------------------------
#define TRY_COUNT 10
#define OK_COUNT 3
#define TRY_TO_READ (6+6*10)

void __fastcall TForm1::PassiveSweepButtonClick(TObject *Sender)
{
	DCB dcb;
	int ret = GetCommState(handle, &dcb);
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;

	DWORD read;
	BYTE buffer[4096];
	AnsiString e;

	Screen->Cursor = crHourGlass;
	bool found = false;
	DWORD try_to_read = 6+6*TRY_COUNT;

	ret = PurgeComm(handle, PURGE_RXCLEAR|PURGE_TXCLEAR);

	for (int b = 11; b  >= 0 && !found; b--) {
		dcb.BaudRate = baudrates[b];
		ret = SetCommState(handle, &dcb);
		if (!ret)
			throw Exception("SetCommState");
		//Memo1->Lines->Add("set baud " + IntToStr(baudrates[b]));

		for (;;) {
			ret = PurgeComm(handle, PURGE_RXCLEAR|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_TXABORT);
			ret = ClearCommError(handle, &port_errors, &comstat);
			if (port_errors == 0)
				break;
		}

		if (comstat.cbInQue > 0) {
			ret = ClearCommError(handle, &port_errors, &comstat);
		}

		DWORD tc0 = GetTickCount();

		for ( ; !found; ) {
			DWORD tc = GetTickCount();
			if (tc > tc0+timeout) {
				//Memo1->Lines->Add("timeout "+IntToStr(comstat.cbInQue) + " " + e);
				break;
			}

			if (!ClearCommError(handle, &port_errors, &comstat))
				throw Exception("ClearCommError()");

			if (port_errors) {
				AnsiString e = "-----";
				if (port_errors&CE_BREAK)
					e[1] = 'b';
				if (port_errors&CE_FRAME)
					e[2] = 'f';
				if (port_errors&CE_OVERRUN)
					e[3] = 'o';
				if (port_errors&CE_RXOVER)
					e[4] = 'O';
				CommStatLabel->Caption = e;
				//Memo1->Lines->Add(e);
				ret = PurgeComm(handle, PURGE_RXCLEAR|PURGE_RXABORT|PURGE_TXCLEAR|PURGE_TXABORT);
				continue;
			}

			if (comstat.cbInQue >= try_to_read) {
				//Memo1->Lines->Add("cbInQue"+IntToStr(comstat.cbInQue));
				DWORD read;
				BYTE reply[TRY_TO_READ];
				if (!ReadFile(handle, reply, try_to_read, &read, NULL))
					throw Exception("read error");

				int ok_count = 0, dbg_ng_bytes = 0;
				for (int i = 0; i+6 < read; )
					if (reply[i] == STX && reply[i+4] == ETX
					 && (reply[i+1]^reply[i+2]^reply[i+3]^reply[i+4]) == reply[i+5]) {
						if (++ok_count >= OK_COUNT) {
							BaudCombo->ItemIndex = b;
							ErrLabel->Font->Color = clNavy;
							ErrLabel->Caption = "SUCCEED";
							found = true;
							break;
						}
						i += 6;
					} else {
						i++;
						ok_count = 0;
						dbg_ng_bytes++;
					}
				break;
			}
		}
	}
	Screen->Cursor = crDefault;
}
//---------------------------------------------------------------------------
void __fastcall TForm1::BatchExecute(TObject *Sender)
{
	DCB dcb;
	GetCommState(handle, &dcb);
	dcb.ByteSize = 8;
	dcb.Parity = NOPARITY;
	dcb.StopBits = ONESTOPBIT;
	int baud_index = 12;

	int last_read, last_displ;
	DWORD last_recv_tc = 0, sec_tc;
	int baud_change = 0;
	int pe_cnt = 0;
	sec_tc = 0;
	abort_flag = false;
	for ( ; !abort_flag; ) {
		Application->ProcessMessages();
		if (handle == INVALID_HANDLE_VALUE)
			continue;

		DWORD tc = GetTickCount();
		if (tc >= sec_tc + 1000) {
			BytesSecLabel->Caption = IntToStr(total_read-last_read);
			CountSecLabel->Caption = IntToStr(total_displ-last_displ);
			last_read = total_read;
			last_displ = total_displ;
			sec_tc = tc;
		}

		if (tc >= last_recv_tc + 100)
			baud_change = 1;

		if (baud_change) {
			baud_index--;
			if (baud_index < 0)
				baud_index = 11;
			dcb.BaudRate = (int)baudrates[baud_index];
			SetCommState(handle, &dcb);
			PurgeComm(handle, PURGE_RXCLEAR|PURGE_TXCLEAR);
			BaudCombo->ItemIndex = baud_index;

			const BYTE cmd[5] = "\x02M1\x03\x7F";	// CONTINUOUS
			DWORD written;
			if (!WriteFile(handle, cmd, 5, &written, NULL))
				throw Exception("WriteFile() error");
			baud_change = 0;
			last_recv_tc = tc;
		}

		if (!ClearCommError(handle, &port_errors, &comstat))
			throw Exception("ClearCommError()");
		CommStatLabel->Caption = stat_str(port_errors);
		if (port_errors != 0) {
			burst_count = 0;
			++pe_cnt;
			//if (pe_cnt >= 3)
			//	baud_change = 1;
		} else if (comstat.cbInQue > 0) {
			DWORD read, toread = comstat.cbInQue;
			if (toread > BURST_SIZE-burst_count)
				toread = BURST_SIZE-burst_count;
			if (!ReadFile(handle, burst_buffer, toread, &read, NULL))
				throw Exception("ReadFile()");
			burst_count += read;
			total_read += read;

			int last_index = -1, last_stx = -1, displ;
			for (int i = 0; i < burst_count; i++) {
				if (burst_buffer[i] == STX) {
					last_stx = i;
					if (i+6 < burst_count
					 && burst_buffer[i+4] == ETX
					 && (burst_buffer[i+1] ^ burst_buffer[i+2] ^ burst_buffer[i+3] ^ burst_buffer[i+4]) == burst_buffer[i+5]) {
						pe_cnt = 0;
						last_index = i;
						total_displ++;
						last_recv_tc = GetTickCount();
						displ = (burst_buffer[i+1]<<16) + (burst_buffer[i+2]<<8) + burst_buffer[i+3];
					}
				}
			}

			if (last_index >= 0) {
				MoveMemory(burst_buffer, burst_buffer+last_index+6, burst_count-(last_index+6));
				burst_count -= last_index+6;
				TotalReadLabel->Caption = IntToStr(total_read);
				TotalDisplLabel->Caption = IntToStr(total_displ);

				DisplEdit->Text = IntToHex(displ, 6);
				if (base_dist > 0.0)
					DisplEdit->Text = DisplEdit->Text +","
					  + AnsiString().sprintf("%8.3fmm", base_dist + range*(displ - 0x100000)/0xAAAAA);

				last_recv_tc = GetTickCount();
			} else if (last_stx >= 0) {
				MoveMemory(burst_buffer, burst_buffer+last_stx, burst_count-last_stx);
				burst_count -= last_stx;
			} else {
				burst_count = 0;
			}
		}
	}
}


void __fastcall TForm1::AbortExecute(TObject *Sender)
{
	abort_flag = true;
}
//---------------------------------------------------------------------------

